package com.jeffknecht.dynadao.api;

import java.lang.annotation.Annotation;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import com.jeffknecht.dynadao.impl.DynaDaoImpl;

/**
 * <p>
 * Provides a dynamic proxy for a DAO interface.
 * </p>
 * 
 * <p>
 * There are some specific rules the DAO interface must follow:
 * </p>
 * 
 * <p>First, the interface may only declare the following methods:
 * <ul>
 * <li>int insert(Object<T>)</li>
 * <li>int update(Object<T>)</li>
 * <li>int delete(Object<T>)</li>
 * <li>List<T> findAll()</li>
 * <li>T findBy...(args...)</li>
 * <li>List<T> findBy...(args...)</li>
 * </ul>
 * </p>
 * 
 * <p>Second, either the interface or each of the find methods
 * must be annotated with the DTOClass annotation, defining
 * the class (<T>) that is returned by the find... methods</p>
 * 
 * <p>Third, every parameter of a find... method must be
 * annotated with a Parameter annotation, defining the property
 * in the DTOClass that is represented by the parameter</p>
 * 
 * <p>Any call to a method that does not follow these guidelines
 * may be met with a nasty runtime exception. You have been warned!</p>
 * 
 * @author jeff.knecht
 *
 */
public class DynaDaoProxy {

	/**
	 * Builds and returns a DynaDao proxy based on the supplied interface.
	 * 
	 * @param intfce The interface to proxy
	 * @param dynaDao The DynaDao instance to pass the calls to.
	 * @return
	 */
	public static <T> T getProxy(Class<T> intfce, DynaDao dynaDao) {
		InvocationHandler handler = new DynaDaoProxyInvocationHandler(intfce, dynaDao);
		return (T) Proxy.newProxyInstance(Thread.currentThread().getContextClassLoader(), new Class<?>[] {intfce}, handler);
	}
	
	private static final class DynaDaoProxyInvocationHandler implements InvocationHandler {
		private DynaDao dynaDao;
		private Class<?> proxyInterface;
		
		private DynaDaoProxyInvocationHandler(Class proxyInterface, DynaDao dynaDao) {
			this.dynaDao = dynaDao;
			this.proxyInterface = proxyInterface;
		}
		
		public Object invoke(Object proxy, Method method, Object[] args)
				throws Throwable {
			String methodName = method.getName();
			
			if (methodName.equals("insert")) {
				return dynaDao.insert(args[0]);
			} else if (methodName.equals("update")) {
				return dynaDao.update(args[0]);
			} else if (methodName.equals("delete")) {
				return dynaDao.delete(args[0]);
			} else if (methodName.equals("findAll")) {
				DTOClass type = method.getAnnotation(DTOClass.class);
				if (type == null) {
					type = proxyInterface.getAnnotation(DTOClass.class);
				}
				return dynaDao.findAll(type.value());
			} else if (methodName.startsWith("find")) {
				DTOClass type = method.getAnnotation(DTOClass.class);
				if (type == null) {
					type = proxyInterface.getAnnotation(DTOClass.class);
				}
				
				HashMap<String, Object> params = new HashMap<String, Object>();
				StringBuilder sb = new StringBuilder();
				
				Annotation[][] paramAnnotations = method.getParameterAnnotations();
				for (int i = 0; i < paramAnnotations.length; i++) {
					for (int j = 0; j < paramAnnotations[i].length; j++) {
						Annotation paramAnnot = paramAnnotations[i][j];
						if (paramAnnot instanceof Parameter) {
							String paramName = ((Parameter) paramAnnot).value();
							if (sb.length() > 0) {
								sb.append(" and ");
							}
							sb.append("{" + paramName + "} = :" + paramName);
							params.put(paramName, args[i]);
						}
					}
				}
			
				if (method.getReturnType().isAssignableFrom(List.class)) {
					return dynaDao.find(type.value(), sb.toString(), params);
				} else {
					return dynaDao.findOne(type.value(), sb.toString(), params);
				}
			} else {
				throw new UnsupportedOperationException("Unsupported method: " + methodName);
			}
		}
		
	}
	
}
